# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flask
import json
import logging
import os
import uuid

from redis import Redis
from rq.command import send_shutdown_command
from rq.worker import Worker

from dcifeeder import routine
from dcifeeder import settings
from dcifeeder.bus import Bus
from dcifeeder.event import upgrade_event
from dcifeeder.logging import configure_feeder_logger
from dcifeeder.workers import (
    RHELWorker,
    QEKNIWorker,
    OpenShiftWorker,
    RHELNoInstallWorker,
)
from dcifeeder.topics.openshift import Openshift, get_supported_openshift_versions
from dcifeeder.topics.openstack import OPENSTACK_TOPICS_LIST

configure_feeder_logger()
logger = logging.getLogger(__name__)


def dump_settings(app):
    symbols = filter(lambda x: x.isupper(), dir(settings))
    values = "\n\t".join(
        (
            "%s: '%s'"
            % (
                sym,
                getattr(settings, sym) if "PASSWORD" not in sym else "*** REDACTED ***",
            )
            for sym in symbols
        )
    )
    logger.info("Dumping settings:\n\t%s" % values)


app = flask.Flask(__name__)
dump_settings(app)


@app.route("/", strict_slashes=False)
def index():
    return flask.Response(
        json.dumps(
            {
                "status": "OK",
                "message": "Distributed CI Feeder as a Service. GET /options to see available options",
            }
        ),  # noqa
        status=200,
        content_type="application/json",
    )


@app.route("/openstack/options", strict_slashes=False)
def options():
    return flask.Response(
        json.dumps(
            {
                "message": "Available call POST /openstack/<topic_name> (ie. POST /openstack/osp16)",
                "topics": list(OPENSTACK_TOPICS_LIST.keys()),
            }
        ),
        status=200,
        content_type="application/json",
    )


@app.route("/openstack/<topic_name>", strict_slashes=False, methods=["POST"])
def sync_openstack(topic_name):
    topic_class = OPENSTACK_TOPICS_LIST.get(topic_name)
    if topic_class is None:
        return flask.Response(
            json.dumps(
                {
                    "message": f"{topic_name} topic does not exists. Please query GET /openstack/options."
                }
            ),
            status=404,
            content_type="application/json",
        )
    topic = topic_class()
    event_id = str(uuid.uuid4())
    logger.info(f"OpenStack Sync - Topic {topic.name} - Event id {event_id}")
    routine.get_jobs_queue("main").enqueue(
        routine.start_feeder, args=(topic, event_id), job_id=event_id
    )
    return flask.Response(
        json.dumps(
            {
                "status": "OK",
                "message": f"Feeder for {topic.name} topic has been started.",
                "event_id": event_id,
                "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
            }
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/openstack/<topic_name>", strict_slashes=False, methods=["PATCH"])
def update_topic_tags(topic_name):
    topic_class = OPENSTACK_TOPICS_LIST.get(topic_name)
    if topic_class is None:
        return flask.Response(
            json.dumps(
                {
                    "message": f"{topic_name} topic does not exists. Please query GET /openstack/options."
                }
            ),
            status=404,
            content_type="application/json",
        )
    topic = topic_class()
    event_id = str(uuid.uuid4())
    logger.info(f"OpenStack Tagging - Topic {topic.name} - Event id {event_id}")
    routine.get_jobs_queue("quick").enqueue(
        routine.update_topic_tags, args=(topic, event_id), job_id=event_id
    )
    return flask.Response(
        json.dumps(
            {
                "status": "OK",
                "message": f"Feeder tags update for {topic.name} has been started.",
                "event_id": event_id,
                "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
            }
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/openshift/<ocp_version>", strict_slashes=False)
@app.route("/ocp/<ocp_version>", strict_slashes=False)
def sync_ocp(ocp_version):
    topic = Openshift(ocp_version)
    event_id = str(uuid.uuid4())
    logger.info(f"OCP Sync - Topic {topic.name} - Event id {event_id}")
    routine.get_jobs_queue("quick").enqueue(
        routine.start_ocp_sync, args=(topic, event_id), job_id=event_id
    )
    return flask.Response(
        json.dumps(
            {
                "status": "OK",
                "message": f"Feeder for openshift {ocp_version} has been started.",
                "event_id": event_id,
                "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
            }
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/synchronize_openshift_components", strict_slashes=False)
def synchronize_openshift_components():
    ocp_versions = get_supported_openshift_versions()
    events = []
    for ocp_version in ocp_versions:
        topic = Openshift(ocp_version)
        event_id = str(uuid.uuid4())
        logger.info(f"OCP Sync - Topic {topic.name} - Event id {event_id}")
        routine.get_jobs_queue("quick").enqueue(
            routine.start_ocp_sync, args=(topic, event_id), job_id=event_id
        )
        events.append((ocp_version, event_id))
    return flask.Response(
        json.dumps(
            [
                {
                    "status": "OK",
                    "message": f"Feeder for openshift {ocp_version} has been started.",
                    "event_id": event_id,
                    "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
                }
                for ocp_version, event_id in events
            ]
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/events", strict_slashes=False, methods=["POST"])
def sync_umb_event():
    event = upgrade_event(flask.request.json)
    bus = Bus(
        workers=[
            RHELWorker(),
            QEKNIWorker(),
            OpenShiftWorker(),
            RHELNoInstallWorker(),
        ]
    )
    event_ids = bus.dispatch(event)
    return flask.Response(
        json.dumps(
            [
                {
                    "status": "OK",
                    "message": "Event propagated",
                    "event_id": event_id,
                    "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
                }
                for event_id in event_ids
            ]
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/logs/<event_id>", strict_slashes=False)
def get_job_log(event_id):
    try:
        with open(os.path.join(settings.LOG_DIR, str(event_id))) as log_fd:
            return flask.Response(log_fd.read(), content_type="text/plain")
    except Exception:
        return flask.Response(None, status=404)


@app.route("/remove_unsupported_components", methods=["POST"])
def remove_unsupported_components():
    event_id = str(uuid.uuid4())
    routine.get_jobs_queue("quick").enqueue(
        routine.remove_unsupported_components, args=(event_id,), job_id=event_id
    )
    return flask.Response(
        json.dumps(
            {
                "status": "OK",
                "message": "Start removing unsupported components.",
                "event_id": event_id,
                "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
            }
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/clean/components_state", methods=["POST"])
def update_components_state():
    event_id = str(uuid.uuid4())
    routine.get_jobs_queue("quick").enqueue(
        routine.update_components_state, args=(event_id,), job_id=event_id
    )
    return flask.Response(
        json.dumps(
            {
                "status": "OK",
                "message": "Update component states based on domain rules.",
                "event_id": event_id,
                "rel": f"http://feeder.distributed-ci.io/logs/{event_id}",
            }
        ),
        status=201,
        content_type="application/json",
    )


@app.route("/admin/workers/shutdown", methods=["PATCH"])
def shutdown_rq_workers() -> flask.Response:
    connection = Redis(host=settings.RQ_REDIS_HOST, port=settings.RQ_REDIS_PORT)
    for worker in Worker.all(connection=connection):
        send_shutdown_command(connection, worker.name)

    return flask.Response(
        json.dumps(
            {
                "status": "Accepted",
                "message": "Workers will shutdown once their current jobs are done.",
            }
        ),
        status=202,
        content_type="application/json",
    )
