# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

logger = logging.getLogger(__name__)


class BaseComponent(object):
    def __init__(self, metadata_url):
        self.data = {}
        self.metadata_url = metadata_url
        self.state = "active"
        self.tags = []
        self.api_accepted_fields = [
            "uid",
            "topic_id",
            "title",
            "message",
            "data",
            "display_name",
            "version",
            "url",
            "type",
            "state",
            "released_at",
            "tags",
        ]

    def update_from_api(self, api_component):
        self.id = api_component.get("id")
        self.version = api_component.get("version")
        self.type = api_component.get("type")
        self.state = api_component.get("state")
        self.released_at = api_component.get("released_at")
        self.uid = api_component.get("uid")
        self.topic_id = api_component.get("topic_id")

        self.etag = api_component.get("etag", None)
        self.title = api_component.get("title", None)
        self.message = api_component.get("message", None)
        self.display_name = api_component.get("display_name", None)
        self.url = api_component.get("url", None)

        api_data = api_component.get("data", None)
        if api_data is not None:
            data = api_data.copy()
            data.update(self.data)
            self.data = data

        tags = api_component.get("tags", None)
        if tags is not None:
            self.tags = list(set(self.tags + tags))

    def serialize(self):
        return dict(self)

    def __str__(self):
        component = {}
        fields = self.api_accepted_fields + ["id", "etag"]
        for field in fields:
            if hasattr(self, field):
                component[field] = getattr(self, field)

        return str(component)

    def __iter__(self):
        for field in self.api_accepted_fields:
            if hasattr(self, field):
                yield (field, getattr(self, field))

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, item, value):
        if item in self.api_accepted_fields + ["id", "etag"]:
            setattr(self, item, value)
        else:
            raise KeyError("invalid key for component: %s" % item)

    def is_healthy(self):
        """is_healthy() is called during the synchronization.

        Use this to stop a sync if component is invalid.
        """
        raise NotImplementedError

    def retrieve_metadata(self, metadata_url):
        """retrieve_metadata() is responsible for populating our component
        object with metadata from remote sources (metadata_url).

        If you need to populate your component with such data, override
        this method instead of doing http requests in the constructor, this
        is much better for everybody's brains when debugging issues.
        """
        raise NotImplementedError

    def update_metadata(self, files_list):
        """Updates the components' metadata contained in the "data" field"""
        pass
