# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
import requests

from dcifeeder.component import BaseComponent


logger = logging.getLogger(__name__)

ten_seconds = 10 * 1000
five_minutes = 5 * 60 * 1000
thirty_minutes = 30 * 60 * 1000


def _get_composeinfo_url(metadata):
    return os.path.join(metadata["url"], "compose/.composeinfo")


def get_composeinfo(metadata):
    composeinfo_url = _get_composeinfo_url(metadata)
    logger.debug("Try to get %s" % composeinfo_url)
    r = requests.get(composeinfo_url)
    r.raise_for_status()


def get_package_file_version(files_list, package, arch):
    package_rex = re.compile(r"^%s-(?P<version>\d.+)\.%s\.rpm$" % (package, arch))
    matches = list(
        filter(
            bool,
            (package_rex.match(file_desc["name"]) for file_desc in files_list["files"]),
        )
    )
    return matches[0].group("version")


class Compose(BaseComponent):
    def __init__(self, metadata):
        self.url = metadata["url"].rstrip("/")
        super(Compose, self).__init__(self.url)
        self.display_name = metadata["display_name"]
        self.version = metadata["version"]
        self.type = metadata["type"]
        self.data = metadata
        self.tags = metadata["tags"]
        self.released_at = metadata["released_at"]

    def retrieve_metadata(self, metadata_url):
        pass

    def is_healthy(self):
        try:
            get_composeinfo(self.data)
            return True
        except Exception:
            return False

    def update_metadata(self, files_list):
        if self.type in ("compose", "compose-noinstall"):
            try:
                kernel_version = get_package_file_version(
                    files_list, "kernel", "x86_64"
                )
                tag = "kernel:%s" % kernel_version
                if tag not in self.tags:
                    self.tags.append(tag)
            except Exception:
                logger.exception(
                    "Error getting kernel version for component %s" % self.display_name
                )
