# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import requests
import yaml
import subprocess

from dcifeeder import settings
from dcifeeder.exceptions import FailureException
from dcifeeder.components import rhos_release_containers

logger = logging.getLogger(__name__)


def copy_image(image, username, password, insecure=False):
    source_url = "%s:%s" % (image["origin"]["repository"], image["origin"]["tag"])
    destination_url = "%s:%s" % (image["dest"]["repository"], image["dest"]["tag"])
    logger.info("Copying image from %s to %s" % (source_url, destination_url))
    r = subprocess.run(
        [
            "skopeo",
            "copy",
            "--dest-tls-verify=%s" % (not insecure),
            "--dest-creds=%s:%s" % (username, password),
            "docker://%s" % source_url,
            "docker://%s" % destination_url,
        ],
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if r.returncode != 0:
        raise FailureException(r.stderr)
    logger.info("... success")


def list_existing_pushed_tags(image, username, password):
    logger.info("Listing existing tags on target registry ...")
    payload = {
        "service": "Docker registry",
        "scope": image["dest"]["tags_list_auth_scope"],
    }
    auth = requests.get(
        image["dest"]["tags_list_auth_url"], auth=(username, password), params=payload
    )
    if auth.status_code == 200 and auth.json().get("token"):
        token = auth.json()["token"]
        headers = {"Authorization": "Bearer %s" % token}
        r = requests.get(image["dest"]["tags_list_url"], headers=headers)
        if r.status_code == 404:
            return []
        elif r.status_code == 200 and r.json() and "tags" in r.json():
            return r.json()["tags"]
        else:
            raise Exception(r.text)
    else:
        raise Exception(auth.text)


def _images_for_puddle(all_images, puddle_rhosp):
    return list(zip(*all_images[puddle_rhosp]))[1]


def _ceph_elts(container_image_prepare):
    """Extract data related to Ceph containers from container-image-prepare
    and make them fit into the dict we're using to mirror all the containers.
    """
    if "ceph-namespace" in container_image_prepare["container-image-prepare"]:
        return {"ns": "ceph-namespace", "img": "ceph-image", "tag": "ceph-tag"}
    elif "ceph_namespace" in container_image_prepare["container-image-prepare"]:
        return {"ns": "ceph_namespace", "img": "ceph_image", "tag": "ceph_tag"}
    else:
        raise FailureException(
            "container-image-prepare missing Ceph container information"
        )


def _ceph_extra(label, container_image_prepare):
    """Extract data related to Ceph "extra" containers from container-image-prepare
    and make them fit into the dict we're using to mirror all the containers.
    """
    if "ceph_%s_image" % label in container_image_prepare["container-image-prepare"]:
        return {
            "namespace": container_image_prepare["container-image-prepare"][
                "ceph_%s_namespace" % label
            ],
            "name": container_image_prepare["container-image-prepare"][
                "ceph_%s_image" % label
            ],
            "tag": container_image_prepare["container-image-prepare"][
                "ceph_%s_tag" % label
            ],
        }

    return None


def _parse_container_image_prepare(container_image_prepare):
    if (
        container_image_prepare["container-image-prepare"]["namespace"]
        == "registry-proxy.engineering.redhat.com/rh-osbs"
    ):
        origin_registry, origin_project = container_image_prepare[
            "container-image-prepare"
        ]["namespace"].split("/")
        dest_project = container_image_prepare["container-image-prepare"][
            "prefix"
        ].split("-")[0]
    else:
        origin_registry, origin_project = container_image_prepare[
            "container-image-prepare"
        ]["namespace"].split("/")
        dest_project = origin_project

    containers_conf = {
        "tag": str(container_image_prepare["container-image-prepare"]["tag"]),
        "puddle": str(container_image_prepare["puddle"]["rhosp"]),
        "origin": {
            "registry": origin_registry,
            "project": origin_project,
            "prefix": container_image_prepare["container-image-prepare"][
                "prefix"
            ].replace("openstack-", ""),
        },
        "dest": {
            "registry_proto": "http" if settings.DEST_REGISTRY_INSECURE else "https",
            "registry": settings.DEST_REGISTRY_URL,
            "project": dest_project,
        },
    }

    ceph = _ceph_elts(container_image_prepare)
    containers_conf["ceph"] = {
        "ceph": {
            "namespace": container_image_prepare["container-image-prepare"][ceph["ns"]],
            "tag": container_image_prepare["container-image-prepare"][ceph["tag"]],
            "name": container_image_prepare["container-image-prepare"][ceph["img"]],
        },
    }

    for extra_label in ("alertmanager", "grafana", "node_exporter", "prometheus"):
        extra_conf = _ceph_extra(extra_label, container_image_prepare)
        if extra_conf:
            containers_conf["ceph"][extra_label] = extra_conf

    return containers_conf


def prepare_image(image_name, container_image_prepare):
    container_image_prepare.update({"name": image_name})

    if container_image_prepare["origin"]["project"]:
        origin_repo = (
            "{origin[registry]}/{origin[project]}/{origin[prefix]}{name}".format(
                **container_image_prepare
            )
        )
    else:
        origin_repo = "{origin[registry]}/{origin[prefix]}{name}".format(
            **container_image_prepare
        )

    return {
        "origin": {"repository": origin_repo, "tag": container_image_prepare["tag"]},
        "dest": {
            "repository": "{dest[registry]}/{dest[project]}/{name}".format(
                **container_image_prepare
            ),
            "tag": container_image_prepare["tag"],
            "tags_list_url": "{dest[registry_proto]}://{dest[registry]}/v2/{dest[project]}/{name}/tags/list".format(
                **container_image_prepare
            ),
            "tags_list_auth_scope": "repository:{dest[project]}/{name}:pull".format(
                **container_image_prepare
            ),
            "tags_list_auth_url": "{dest[registry_proto]}://{dest[registry]}/auth".format(
                **container_image_prepare
            ),
        },
    }


def prepare_ceph_images(container_image_prepare):
    ceph_images = []

    for label, ceph_image in container_image_prepare["ceph"].items():
        args = {"ceph": ceph_image, "dest": container_image_prepare["dest"]}

        ceph_images.append(
            {
                "origin": {
                    "repository": "{ceph[namespace]}/{ceph[name]}".format(**args),
                    "tag": args["ceph"]["tag"],
                },
                "dest": {
                    "repository": "{dest[registry]}/{dest[project]}/{ceph[name]}".format(
                        **args
                    ),
                    "tag": container_image_prepare["tag"],
                    "tags_list_url": "{dest[registry_proto]}://{dest[registry]}/v2/{dest[project]}/{ceph[name]}/tags/list".format(
                        **args
                    ),
                    "tags_list_auth_scope": "repository:{dest[project]}/{ceph[name]}:pull".format(
                        **args
                    ),
                    "tags_list_auth_url": "{dest[registry_proto]}://{dest[registry]}/auth".format(
                        **args
                    ),
                },
            }
        )
    return ceph_images


def _sync_image(image):
    if image["dest"]["tag"] in list_existing_pushed_tags(
        image, settings.DEST_REGISTRY_USER, settings.DEST_REGISTRY_PASSWORD
    ):
        logger.info("{dest[repository]}:{dest[tag]} already pushed".format(**image))
    else:
        copy_image(
            image,
            settings.DEST_REGISTRY_USER,
            settings.DEST_REGISTRY_PASSWORD,
            settings.DEST_REGISTRY_INSECURE,
        )


def _sync_images(overcloud_container_image_prepare):
    container_image_prepare_raw = yaml.safe_load(overcloud_container_image_prepare)
    container_image_prepare = _parse_container_image_prepare(
        container_image_prepare_raw
    )

    images = _images_for_puddle(
        rhos_release_containers.CONTAINERS, container_image_prepare["puddle"]
    )
    uploaded_images = []

    for image_name in images:
        image = prepare_image(image_name, container_image_prepare)
        _sync_image(image)
        uploaded_images.append("{dest[repository]}:{dest[tag]}".format(**image))

    ceph_images = prepare_ceph_images(container_image_prepare)
    for ceph_image in ceph_images:
        _sync_image(ceph_image)
        uploaded_images.append("{dest[repository]}:{dest[tag]}".format(**ceph_image))

    return uploaded_images


def mirror_images(images_list_loc):
    r = requests.get(images_list_loc)
    images_list = _sync_images(r.text)
    return yaml.dump(images_list, explicit_start=True, default_flow_style=False)
