# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import requests

from dcifeeder.component import BaseComponent
from dcifeeder.exceptions import FailureException
from urllib.parse import urlparse

logger = logging.getLogger(__name__)


def is_not_under_embargo(component):
    r = requests.get("%s/EMBARGOED" % component.url)
    if r.status_code == 200:
        logger.error("EMBARGOED: Puddle is under embargo")
        return False
    return True


def is_status_file_correct(component):
    """Due to some changes in the build tool starting with OSP15, we're testing both status.txt
    and STATUS files"""

    r = requests.get("%s/status.txt" % component.url)
    if r.status_code == 200 and r.text.strip() == "COMPLETE":
        return True

    r = requests.get("%s/STATUS" % component.url)
    if r.status_code == 200 and r.text.strip() == "FINISHED":
        return True

    logger.info("Compose status is not COMPLETE")
    return False


def is_overcloud_container_present(component):
    r = requests.get(component.overcloud_container_image_prepare_url)
    if r.status_code != 200:
        logger.error("overcloud_container_image_prepare.yaml: Missing file")
        return False
    return True


class OSPFullComposeWithContainers(BaseComponent):
    def __init__(self, metadata_url):
        super(OSPFullComposeWithContainers, self).__init__(metadata_url)
        self.type = "compose"

    def retrieve_metadata(self, metadata_url):
        if not metadata_url.endswith("COMPOSE_ID"):
            raise FailureException("Can't parse metadata file of unknown type.")

        r = requests.get(metadata_url)
        if r.status_code != 200:
            raise FailureException(
                "Error fetching component metadata: %s - Status code: %d"
                % (metadata_url, r.status_code)
            )

        self.display_name = r.text
        self.version = (
            self.display_name.split("-", 1)[1] if "-" in self.display_name else ""
        )
        product_url = "/".join(r.url.split("/")[:-2])
        self.url = "/".join((product_url, r.text))
        self.data = {
            "path": urlparse(self.url).path,
            "version": self.version,
            "repo_name": self.display_name,
        }

    @property
    def overcloud_container_image_prepare_url(self):
        return "%s/container_image_prepare.yaml" % self.url

    def is_healthy(self):
        return (
            is_not_under_embargo(self)
            and is_status_file_correct(self)
            and is_overcloud_container_present(self)
        )
