import logging
from datetime import datetime

logger = logging.getLogger(__name__)


def upgrade_event(event):
    """
    In-place upgrade legacy message formats to allow transision from new to
    old event formats.
    """
    if "headers" not in event:
        event = {"msg": event, "headers": {}}
    return event


def is_a_rhel_event(event):
    try:
        tag = event["msg"]["tag"]
        if tag not in ["nightly", "candidate", "milestone"]:
            return False
        compose = event["msg"]["compose"]["compose_info"]["payload"]["compose"]
        datetime.strptime(compose["date"], "%Y%m%d")
        compose_id = compose["id"].lower()
        return compose_id.startswith("rhel-")
    except Exception:
        return False


def is_a_pungi_event(event):
    try:
        release_short = event["msg"]["release_short"]
        is_rhel = release_short == "RHEL"

        status = event["msg"]["status"]
        is_finished = status == "FINISHED"

        release_type = event["msg"]["release_type"]
        is_ga = release_type == "ga"

        compose_type = event["msg"]["compose_type"]
        type_is_correct = compose_type in ["nightly", "candidate", "milestone"]
        datetime.strptime(event["msg"]["compose_date"], "%Y%m%d")
        return all([is_rhel, is_finished, is_ga, type_is_correct])
    except Exception:
        return False


def _get_supported_arches(major_version):
    if major_version == "7":
        return ["x86_64", "ppc64le"]
    return ["x86_64", "s390x", "ppc64", "ppc64le"]


def extract_metadata_from_cts_event(event):
    payload = event["msg"]["compose"]["compose_info"]["payload"]
    release = payload["release"]
    version = release["version"]
    versions_digits = version.split(".")
    major = versions_digits[0] if len(versions_digits) > 0 else None
    minor = versions_digits[1] if len(versions_digits) > 1 else None
    patch = versions_digits[2] if len(versions_digits) > 2 else None
    compose = payload["compose"]

    tags = set()
    tags.add(event["msg"]["tag"])
    for tag in event["msg"]["compose"]["tags"]:
        if tag in ["nightly", "candidate", "milestone"]:
            tags.add(tag)

    compose_url = event["msg"]["compose"]["compose_url"]
    released_at = datetime.strptime(compose["date"], "%Y%m%d")
    return {
        "product": release["short"],
        "url": compose_url,
        "version": compose["id"].replace("RHEL-", ""),
        "major": major,
        "minor": minor,
        "patch": patch,
        "display_name": compose["id"],
        "arches": _get_supported_arches(major),
        "tags": sorted(list(tags)),
        "type": "compose",
        "released_at": datetime.strftime(released_at, "%Y-%m-%dT%H:%M:%S.%f"),
    }


def extract_metadata_from_pungi_event(event):
    version = event["msg"]["release_version"]
    versions_digits = version.split(".")
    major = versions_digits[0] if len(versions_digits) > 0 else None
    minor = versions_digits[1] if len(versions_digits) > 1 else None
    patch = versions_digits[2] if len(versions_digits) > 2 else None
    compose_type = event["msg"]["compose_type"]
    compose_id = event["msg"]["compose_id"]
    released_at = datetime.strptime(event["msg"]["compose_date"], "%Y%m%d")
    return {
        "product": event["msg"]["release_short"],
        "url": event["msg"]["location"].replace("/compose", "/"),
        "version": compose_id.replace("RHEL-", ""),
        "major": major,
        "minor": minor,
        "patch": patch,
        "display_name": compose_id,
        "arches": _get_supported_arches(major),
        "tags": [compose_type],
        "type": "compose-noinstall",
        "released_at": datetime.strftime(released_at, "%Y-%m-%dT%H:%M:%S.%f"),
    }


def is_a_kni_baremetal_event(event):
    try:
        team = event["msg"]["contact"]["team"]
        return team.lower() == "openshift kni qe team"
    except Exception:
        return False


def extract_metadata_from_kni_event(event):
    msg = event["msg"]
    version = msg["artifact"]["release"].split("-")[0]
    major, minor, patch = version.split(".")
    return {
        "test": msg["test"],
        "ocp": {
            "id": msg["artifact"]["id"],
            "product_name": msg["artifact"]["name"],
            "topic_name": "OCP-%s.%s" % (major, minor),
            "pull_from": msg["artifact"]["pullFrom"],
            "version": version,
            "major": major,
            "minor": minor,
            "patch": patch,
        },
    }


def is_dci_openshift_event(event):
    try:
        return event["headers"]["EVENT_TYPE"] == "DCI_OPENSHIFT_EVENT"
    except Exception:
        return False


def extract_metadata_from_dci_openshift_event(event):
    msg = event["msg"]
    version = msg["version"]
    major, minor, patch = version.split(".")
    return {
        "product": "OpenShift",
        "tags": msg["tags"],
        "release_txt_url": msg["release_txt_url"],
        "version": version,
        "major": major,
        "minor": minor,
        "patch": patch,
    }
