# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dciclient.v1.api.component as dci_component
import dciclient.v1.api.context as dci_context
import dciclient.v1.api.topic as dci_topic
import dciclient.v1.api.product as dci_product
import logging
import pipes
import shlex

from dcifeeder.utils.sort import sort_by_newest_first
from dcifeeder import settings
from dcifeeder.exceptions import FailureException, ResourceNotFound
from functools import wraps
from subprocess import check_output, STDOUT, CalledProcessError

logger = logging.getLogger(__name__)


def contextify(f):
    @wraps(f)
    def func(*args, **kwargs):
        if settings.DCI_CLIENT_ID is None or settings.DCI_API_SECRET is None:
            context = dci_context.build_dci_context(
                settings.DCI_CS_URL, settings.DCI_LOGIN, settings.DCI_PASSWORD
            )
            logger.warning("API credentials not set, fallback on user credentials.")
        else:
            context = dci_context.build_signature_context(
                settings.DCI_CS_URL, settings.DCI_CLIENT_ID, settings.DCI_API_SECRET
            )
        return f(context, *args, **kwargs)

    return func


def run_command(command):
    command_str = build_command_string(command)
    logger.debug("Running command: '%s'" % command_str)
    try:
        stdout = check_output(command, stderr=STDOUT)
    except CalledProcessError as e:
        logger.error(
            "Command '%s' exited with return code: %d" % (command_str, e.returncode)
        )
        logger.error("Output was: %s" % e.output)
        raise FailureException(
            "An error occured while running: %s. "
            "Return code: %d" % (command, e.returncode)
        )

    logger.debug("Command `%s` succeeded." % command_str)
    logger.debug("Output was: %s" % stdout)
    return stdout


def build_command_string(command):
    """
    Compat function around shlex.join()
    This could all be replaced by shlex.join() but it was only introduced in Python 3.8
    Before this shlex.quote() is the correct one to use but only introduced in Python 3.3
    Fallback to using pipex.quote() which is supposed to be deprecated since Python 2.7
    """
    if hasattr(shlex, "join"):
        return shlex.join(command)
    elif hasattr(shlex, "quote"):
        return " ".join((shlex.quote(arg) for arg in command))
    else:
        return " ".join((pipes.quote(arg) for arg in command))


def build_remote_dir(host, path):
    remote_command = build_command_string(("mkdir", "-p", path))
    run_command(("ssh", host, remote_command))


@contextify
def get_topic_from_name(context, name):
    r = dci_topic.list(context, where="name:" + name, limit=1, offset=0)
    r.raise_for_status()
    topics = r.json()["topics"]
    if len(topics) == 0:
        raise ResourceNotFound("Failed getting topic %s: %s" % (name, r.content))
    return topics[0]


@contextify
def get_or_create_topic_from_name(context, name, product_name):
    created = False
    try:
        return get_topic_from_name(name), created
    except ResourceNotFound:
        r = dci_product.list(context, where="name:" + product_name, limit=1, offset=0)
        r.raise_for_status()
        products = r.json()["products"]
        if len(products) == 0:
            raise ResourceNotFound(
                "Failed getting product %s: %s" % (product_name, r.content)
            )
        if len(products[0]["topics"]) == 0:
            raise ResourceNotFound("Product %s has no associated topic." % product_name)
        latest_topic = sort_by_newest_first(products[0]["topics"])[0]
        new_topic = {
            key: value
            for key, value in latest_topic.items()
            if key
            not in [
                "name",
                "etag",
                "id",
                "created_at",
                "updated_at",
                "next_topic_id",
                "component_types_optional",
            ]
        }
        new_topic["name"] = name
        new_topic["next_topic_id"] = latest_topic["id"]
        r = dci_topic.create(context, **new_topic)
        r.raise_for_status()
        created = True
        return get_topic_from_name(name), created


@contextify
def get_or_create_inactive_component(context, topic, component):
    """Gets or create an inactive component.

    First tries to find the component by name with state "inactive".
    If not found, creates a new component with state==inactive.
    """
    logger.info(
        "%s: Get or create inactive component: %s (%s)"
        % (topic.name, component.display_name, component.type)
    )

    api_component = get_componment_by_name_type_and_state(topic, component, "inactive")

    if not api_component:
        component.state = "inactive"
        component.topic_id = topic.id
        r = dci_component.create_v2(context, **component.serialize())
        r.raise_for_status()
        api_component = r.json()["component"]

    component.update_from_api(api_component)

    return component


@contextify
def get_or_create_component(context, topic, component):
    """Gets or create a component.

    First tries to find the component by name
    If not found, creates a new component with state inactive.
    """
    logger.info(
        "%s: Get or create component: %s (%s)"
        % (topic.name, component.display_name, component.type)
    )

    get_component = dci_topic.list_components(
        context,
        topic.id,
        sort="-created_at",
        limit=1,
        offset=0,
        where="display_name:%s,type:%s,version:%s"
        % (component["display_name"], component["type"], component["version"]),
    )
    get_component.raise_for_status()
    api_components = get_component.json()["components"]
    api_component = api_components[0] if api_components else None

    if not api_component:
        component.state = "inactive"
        component.topic_id = topic.id
        r = dci_component.create_v2(context, **component.serialize())
        r.raise_for_status()
        api_component = r.json()["component"]

    component.update_from_api(api_component)

    return component


@contextify
def get_component(context, component_id):
    r = dci_component.get(context, component_id)
    r.raise_for_status()
    return r.json()["component"]


@contextify
def get_componment_by_name_type_and_state(context, topic, component, state):
    r = dci_topic.list_components(
        context,
        topic.id,
        sort="-created_at",
        limit=1,
        offset=0,
        where="display_name:%s,type:%s,state:%s"
        % (component["display_name"], component["type"], state),
    )
    r.raise_for_status()
    components = r.json()["components"]
    if components:
        return components[0]
    return None


@contextify
def get_topic(context, topic_id):
    r = dci_topic.get(context, topic_id)
    r.raise_for_status()
    return r.json()["topic"]


@contextify
def update_component(context, component, data):
    data["etag"] = component["etag"]
    logger.info("Update component: {}".format(component["id"]))
    logger.info("Data: {}".format(data))
    r = dci_component.update(context, component["id"], **data)
    if r.status_code == 200:
        logger.info("Component succesfully updated: {}".format(component["id"]))
        updated_component = r.json()["component"]
        component["etag"] = updated_component["etag"]
    else:
        msg = "Component update failed: {}. Status code {}: {}".format(
            component["id"], r.status_code, r.text
        )
        raise FailureException(msg)


@contextify
def update_component_type_on_topic(context, topic, component):
    component_types = topic.component_types
    component_type = component["type"]
    if component_type not in component_types:
        logger.info(
            "Adding component type %s to %s."
            % (
                component_type,
                topic.name,
            )
        )
        component_types.append(component_type)
        r = dci_topic.update(
            context, topic.id, etag=topic.etag, component_types=component_types
        )
        r.raise_for_status()


@contextify
def get_products(context, **kwargs):
    r = dci_product.list(context, **kwargs)
    r.raise_for_status()
    return r.json()["products"]


def _build_where(kwargs):
    tags = ["tags:%s" % t for t in kwargs.pop("tags", [])]
    kwargs = ["%s:%s" % (key, value) for key, value in kwargs.items()]
    return ",".join(sorted(kwargs + tags))


@contextify
def get_components(
    context, topic_id, sort="-created_at", limit=100, offset=0, **kwargs
):
    r = dci_topic.list_components(
        context,
        id=topic_id,
        sort=sort,
        limit=limit,
        offset=offset,
        where=_build_where(kwargs),
    )
    r.raise_for_status()
    return r.json()["components"]
