# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from redis import Redis
import rq

from dcifeeder import settings
from dcifeeder import feeder
from dcifeeder.components import containers
from dcifeeder import clean
from dcifeeder import notify
from dcifeeder import s3
from dcifeeder.logging import configure_worker_logger

logger = logging.getLogger(__name__)

LOCK_IO_ERROR = BlockingIOError


def get_jobs_queue(name: str) -> rq.Queue:
    return rq.Queue(
        name=settings.RQ_QUEUES[name],
        default_timeout=settings.RQ_JOBS_TIMEOUT,
        connection=Redis(host=settings.RQ_REDIS_HOST, port=settings.RQ_REDIS_PORT),
    )


def process_component(topic, component):
    component.retrieve_metadata(component.metadata_url)
    logger.info(
        f"Processing Component: {component.display_name} - Type: {component.type}"
    )
    if not component.is_healthy():
        logger.warning("Component is not healthy. Skipping it.")
        return

    # If there's an "active" component for this topic/name/type, don't sync again
    existing_component = feeder.get_componment_by_name_type_and_state(
        topic, component, "active"
    )
    if existing_component:
        logger.info(
            f"{topic.name}: An active component already exists for "
            f"{existing_component['display_name']} ({existing_component['type']}) {existing_component['id']} - Skipping sync."
        )
        return

    component = feeder.get_or_create_inactive_component(topic, component)
    logger.info(f"Component {component.display_name} has id {component.id}")
    try:
        s3.sync_compose(component)
        images_list_content = containers.mirror_images(
            component.overcloud_container_image_prepare_url
        )
        s3.add_images_list(component, images_list_content)
        update_component_on_api(component)
        feeder.update_component_type_on_topic(topic, component)
        send_notification(topic, component)
    except Exception as e:
        logger.error("Failure during component sync for %s" % component.display_name)
        logger.exception("Error was: %s" % e)
        raise


def start_feeder(topic, event_id):
    configure_worker_logger(event_id)

    try:
        logger.info("Topic: %s" % topic.name)
        for component in topic.components:
            process_component(topic, component)
    except Exception:
        logger.exception(f"Feeder has failed for topic {topic.name}")
        raise


def update_component_on_api(component):
    logger.info(
        "Processing Component: %s - Type: %s" % (component.display_name, component.type)
    )
    try:
        feeder.update_component(
            component,
            {"state": "active", "data": component.data, "tags": component.tags},
        )
    except Exception as e:
        logger.error("Failure during component update for %s" % component.display_name)
        logger.exception("Error was: %s" % e)
        raise


def send_notification(topic, component):
    notify.send_message_on_google_chat(
        "Component %s for topic %s has been successfully uploaded to DCI https://www.distributed-ci.io/topics/%s/components/%s"
        % (component.display_name, topic.name, topic.id, component.id)
    )


def start_ocp_sync(topic, event_id):
    configure_worker_logger(event_id)
    try:
        logger.info("Starting sync for topic: %s" % topic.name)
        for component in topic.components:
            component = feeder.get_or_create_component(topic, component)
            if component.state == "active":
                logger.info(
                    "%s: An active component already exists for %s (%s). Skipping sync."
                    % (topic.name, component.display_name, component.type)
                )
                continue
            if not component.is_healthy():
                logger.info(
                    f"Component {component.display_name} is not healthy, skipping creation and sync."
                )
                continue
            update_component_on_api(component)
            feeder.update_component_type_on_topic(topic, component)
            send_notification(topic, component)
    except Exception:
        logger.exception("Sync for topic %s has failed" % topic.name)
        logger.error("Status: FAILURE")
        raise


def start_rhel_sync(topic, event_id):
    configure_worker_logger(event_id)
    try:
        logger.info("Starting sync for topic: %s" % topic.name)
        for component in topic.components:
            if not component.is_healthy():
                logger.info(
                    f"Component {component.display_name} is not healthy, skipping creation and sync."
                )
                return
            component = feeder.get_or_create_component(topic, component)
            s3.sync_compose(component)
            update_component_on_api(component)
            feeder.update_component_type_on_topic(topic, component)
            send_notification(topic, component)
    except Exception:
        logger.exception("Sync for topic %s has failed" % topic.name)
        logger.error("Status: FAILURE")
        raise


def remove_unsupported_components(event_id):
    configure_worker_logger(event_id)
    try:
        components_ids = s3.list_components_ids()
        days = settings.SOFT_DELETE_EXPIRATION_DAYS
        for component_id in components_ids:
            component = feeder.get_component(component_id)
            if component is None:
                logger.error(
                    f"Component {component_id} doesn't exist on the control server, but the prefix is present in s3 bucket."
                )
            logger.debug(f"Checking {component['name']} {component['id']}")
            if clean.component_is_inactive_for_more_than_X_days(component, days=days):
                logger.info(
                    f"Component {component['id']} is inactive for more than {days} days. Removing..."
                )
                s3.remove_from_s3(component)
    except Exception as e:
        logger.error("An error occured when removing unsupported components.")
        logger.exception(e)
        raise


def update_topic_tags(topic, event_id):
    configure_worker_logger(event_id)
    # FIXME: does it need a lock ?
    topic.update_tags()


def _rhel_rule(active_components):
    return clean.not_in_latest_x_components_per_type_and_tags(
        active_components,
        ordered_tags=["milestone", "candidate", "nightly"],
        max_nb_components=2,
    )


def _openstack_rule(active_components):
    return clean.not_in_latest_x_components_per_type_and_tags(
        active_components, ordered_tags=[], max_nb_components=1
    )


def update_components_state(event_id):
    configure_worker_logger(event_id)
    logger.info("Update components state")
    try:
        components_to_be_made_inactive = []
        for product in feeder.get_products():
            if product["label"] == "OPENSHIFT":
                continue
            if product["label"] == "RHEL":
                apply_rule = _rhel_rule
            if product["label"] == "OPENSTACK":
                apply_rule = _openstack_rule
            for topic in product["topics"]:
                active_components = feeder.get_components(topic["id"], state="active")
                components_to_be_made_inactive += apply_rule(active_components)
        for component in components_to_be_made_inactive:
            c = feeder.get_component(component_id=component["id"])
            feeder.update_component(c, {"state": "inactive"})
        logger.info("Update components state is finished")
    except Exception as e:
        logger.error("Unable to update components state")
        logger.exception(e)
        raise


def _extract_sha256_from_component(component):
    try:
        return component["data"]["digest"][0].replace("sha256:", "")
    except Exception:
        return None


def _add_qe_ok_tag(component):
    tags = set(component["tags"])
    tags.add("qe-ok")
    feeder.update_component(component, {"tags": list(tags)})


def tag_ocp_component(metadata, event_id):
    configure_worker_logger(event_id)
    version = metadata["ocp"]["version"]
    logger.info("Tag OpenShift %s component with kni-ok tag." % version)
    try:
        if metadata["test"]["result"] != "passed":
            logger.info("kni test are not green. skipping.")
            return
        kni_sha_256 = metadata["ocp"]["id"]
        topic = feeder.get_topic_from_name(metadata["ocp"]["topic_name"])
        component_name = "%s*" % version
        components = feeder.get_components(
            topic_id=topic["id"], state="active", name=component_name
        )
        component_to_tag = None
        for component in components:
            component_sha256 = _extract_sha256_from_component(component)
            if component_sha256 is not None and component_sha256 == kni_sha_256:
                component_to_tag = component
        if component_to_tag:
            _add_qe_ok_tag(component_to_tag)
        else:
            logger.warning("Cannot find OpenShift %s on dci-control-server" % version)
    except Exception as e:
        logger.error("Unable to tag OpenShift %s component." % version)
        logger.exception(e)
        raise
