# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from dataclasses import dataclass, field
import os
import sys
import time
import urllib3
import logging
import tempfile
import gc
import hashlib
import json
import re
import threading
import queue

from base64 import b64decode
from urllib.parse import urlparse

from dcifeeder import settings
from dcifeeder import sync
from dcifeeder.feeder import run_command
import boto3
import botocore.client
from botocore.config import Config
from botocore.exceptions import ClientError
from tenacity import retry, stop_after_attempt


logger = logging.getLogger(__name__)

COPYRIGHT = """
copyright:
  Copyright © 2019-2023 Red Hat.
  Licensed under the Apache License, Version 2.0
"""


def _get_include_list_type(rsync_sourcepath):
    for product in [
        "openstack",
        "rhel-9",
        "rhel-8",
        "rhel-7",
    ]:
        if product in rsync_sourcepath.lower():
            return product


def _get_rsync_sourcepath(compose_url):
    compose_url = urlparse(compose_url)
    compose_path = compose_url.path.rstrip("/")[1:]
    if "openstack" in compose_path.lower():
        base_folder = ""
    else:
        base_folder = "redhat/"
    return "%s::%s%s/compose/" % (compose_url.netloc, base_folder, compose_path)


def parse_arguments(arguments, env_variables):
    parser = argparse.ArgumentParser(
        usage="s3-sync [OPTIONS]",
        description="Synchronize a compose to s3",
        epilog=COPYRIGHT,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--compose-url",
        dest="compose_url",
    )
    parser.add_argument(
        "--include-list-type",
        dest="include_list_type",
    )
    parser.add_argument(
        "--bucket",
        dest="bucket",
    )
    parser.add_argument("--prefix", dest="prefix")
    parser.add_argument(
        "--aws-endpoint-url",
        dest="aws_endpoint_url",
        default=env_variables.get("S3_ENDPOINT_URL"),
    )
    parser.add_argument(
        "--aws-access-key-id",
        dest="aws_access_key_id",
        default=env_variables.get("AWS_ACCESS_KEY_ID"),
    )
    parser.add_argument(
        "--aws-secret-access-key",
        dest="aws_secret_access_key",
        default=env_variables.get("AWS_SECRET_ACCESS_KEY"),
    )
    parser.add_argument(
        "--aws-region",
        dest="aws_region",
        default=env_variables.get("AWS_REGION"),
    )
    parser.add_argument(
        "--aws-signature-version",
        dest="aws_signature_version",
        default=env_variables.get("AWS_SIGNATURE_VERSION"),
    )
    parsed_arguments = parser.parse_args(arguments)
    rsync_sourcepath = _get_rsync_sourcepath(parsed_arguments.compose_url)
    parsed_arguments.rsync_sourcepath = rsync_sourcepath
    parsed_arguments.include_list_type = _get_include_list_type(rsync_sourcepath)
    if parsed_arguments.include_list_type is None:
        print("Please specify an include list type (--include-list-type)")
        sys.exit(2)
    return vars(parsed_arguments)


def build_rsync_commandline(component):
    rsync_sourcepath = _get_rsync_sourcepath(component.url)
    list_type = _get_include_list_type(rsync_sourcepath)
    include_list_mapping = {
        "rhel-7": sync.RHEL_7_RSYNC_LIST,
        "rhel-8": sync.RHEL_8_or_9_RSYNC_LIST,
        "rhel-9": sync.RHEL_8_or_9_RSYNC_LIST,
        "openstack": sync.OPENSTACK_RSYNC_LIST,
    }
    command = ["rsync", "--recursive", "--list-only"]
    include_list = include_list_mapping.get(list_type)
    if include_list:
        for element in include_list:
            command.extend(["--include", element])
        command.extend(["--exclude", "*"])
    command.extend([rsync_sourcepath])
    return command


def create_s3_client(payload: dict) -> botocore.client.BaseClient:
    s3_config = Config(max_pool_connections=50)

    if payload["aws_region"]:
        s3_config.merge(Config(region_name=payload["aws_region"]))
    if payload["aws_signature_version"]:
        s3_config.merge(Config(signature_version=payload["aws_signature_version"]))
    session = boto3.session.Session()
    return session.client(
        "s3",
        endpoint_url=payload["aws_endpoint_url"] or None,
        aws_access_key_id=payload["aws_access_key_id"],
        aws_secret_access_key=payload["aws_secret_access_key"],
        config=s3_config,
    )


def create_bucket_if_not_exists(s3_config) -> None:
    s3_client = create_s3_client(s3_config)
    try:
        s3_client.head_bucket(Bucket=s3_config["bucket"])
    except ClientError as e:
        logger.warning(e.response)
        if int(e.response["Error"]["Code"]) == 404:
            try:
                s3_client.create_bucket(Bucket=s3_config["bucket"])
            except ClientError as exc:
                raise Exception(
                    "Error while creating bucket '%s': %s" % (s3_config["bucket"], exc)
                )
        else:
            raise Exception(
                "Error while creating bucket '%s': %s" % (s3_config["bucket"], str(e))
            )


@dataclass(order=True)
class S3Upload:
    filesize: int
    filename: str
    temp_fd: tempfile.SpooledTemporaryFile = field(default=None)
    sha: str = field(default=None)
    retries_left_down: int = field(default=3)
    retries_left_up: int = field(default=3)


def register_dci_file(dci_files_list: dict, upload: S3Upload) -> None:
    dci_files_list["files"].append(
        {
            "name": os.path.basename(upload.filename),
            "path": os.path.dirname(upload.filename),
            "size": upload.filesize,
            "sha256": upload.sha,
        }
    )


def _get_s3_object_sha(
    s3: botocore.client.BaseClient, payload, upload: S3Upload
):  # TODO: >=py310  -> None | str:
    try:
        obj_head = s3.head_object(
            Bucket=payload["bucket"],
            Key=os.path.join(payload["prefix"], upload.filename),
            ChecksumMode="ENABLED",
        )
        if upload.filesize != obj_head["ContentLength"]:
            return None
        if "ChecksumSHA256" not in obj_head:
            return None
        return b64decode(obj_head["ChecksumSHA256"]).hex()
    except ClientError:
        return None


@retry(stop=stop_after_attempt(3))
def _get_http_file(
    s3: botocore.client.BaseClient,
    payload,
    upload: S3Upload,
) -> bool:
    file_url = os.path.join(payload["compose_url"], "compose", upload.filename)

    upload.sha = _get_s3_object_sha(s3, payload, upload)
    if upload.sha:
        return False

    filehash = hashlib.sha256()
    upload.temp_fd = tempfile.SpooledTemporaryFile(
        mode="w+b",
        max_size=1024 * 1024,
    )
    file_request = payload["http"].request("GET", file_url, preload_content=False)
    if file_request.status != 200:
        raise FileNotFoundError()
    for chunk in file_request.stream(1024 * 1024):
        upload.temp_fd.write(chunk)
        filehash.update(chunk)
    upload.temp_fd.seek(0)
    upload.sha = filehash.hexdigest()

    return True


def _get_http_file_worker(
    s3_config: dict,
    payload: dict,
    download_queue: queue.Queue,
    upload_queue: queue.Queue,
    stop_event: threading.Event,
    fail_event: threading.Event,
) -> None:
    s3_client = create_s3_client(s3_config)

    while not stop_event.is_set() and not fail_event.is_set():
        try:
            upload = download_queue.get_nowait()

            logger.debug(f">> Downloading {upload.filename} - {upload.filesize}")
            if _get_http_file(s3_client, payload, upload):
                upload_queue.put(upload)
                logger.debug(
                    f">> Download done for {upload.filename} - "
                    f"Dls: {download_queue.qsize()} / Ups: {upload_queue.qsize()}"
                )
            else:
                register_dci_file(payload["dci_files_list"], upload)
        except queue.Empty:
            logger.debug("Download queue was empty")
            time.sleep(1)
        except Exception:
            try:
                upload.temp_fd.close()
                del upload.temp_fd
            except Exception:
                pass
            logger.exception(f"Download failed for {upload.filename} - Aborting.")
            download_queue.task_done()
            fail_event.set()
        else:
            download_queue.task_done()

    if fail_event.is_set():
        try:
            while True:
                download_queue.task_done()
        except Exception:
            pass


@retry(stop=stop_after_attempt(3))
def _push_to_s3(s3, payload: dict, upload: S3Upload):
    s3_key = os.path.join(payload["prefix"], upload.filename)

    logger.debug("<< Upload starts for %s" % upload.filename)
    result = s3.put_object(
        Bucket=payload["bucket"],
        Key=s3_key,
        Body=upload.temp_fd,
        ChecksumAlgorithm="sha256",
    )
    logger.debug(f"<< Upload done for {upload.filename}")
    upload.temp_fd.close()

    if result["ResponseMetadata"]["HTTPStatusCode"] != 200:
        raise Exception(
            "Bad status code for Boto3 upload: %d - %s"
            % (result["ResponseMetadata"]["HTTPStatusCode"], result["ResponseMetadata"])
        )

    shaboto = (
        b64decode(result["ChecksumSHA256"]).hex()
        if "ChecksumSHA256" in result
        else None
    )
    if shaboto and upload.sha != shaboto:
        logger.warning(
            "Checksum mismatch for file %s: local=%s, boto=%s"
            % (s3_key, upload.sha, shaboto)
        )


def _push_to_s3_worker(
    s3_config,
    payload: dict,
    upload_queue: queue.Queue,
    stop_event: threading.Event,
    fail_event: threading.Event,
):
    s3_client = create_s3_client(s3_config)
    while not stop_event.is_set() and not fail_event.is_set():
        try:
            upload = upload_queue.get(timeout=1)

            _push_to_s3(s3_client, payload, upload)

            logger.debug("< All done for %s" % upload.filename)
        except queue.Empty:
            pass
        except Exception:
            try:
                upload.temp_fd.close()
            except Exception:
                pass
            upload_queue.task_done()
            fail_event.set()
            logger.exception(f"Upload failed for {upload.filename} - Aborting.")
        else:
            register_dci_file(payload["dci_files_list"], upload)
            upload_queue.task_done()
        finally:
            gc.collect()

    if fail_event.is_set():
        try:
            while True:
                upload_queue.task_done()
        except Exception:
            pass


def sync_compose_on_s3(s3_config, component) -> None:
    logger.info(f"Sync component to S3: {component.display_name} - {component.id}")

    rsync_command = build_rsync_commandline(component)
    rsync_files_list = run_command(rsync_command)

    logging.getLogger("botocore").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    download_queue = queue.Queue()
    upload_queue = queue.Queue(maxsize=50)
    download_done = threading.Event()
    upload_done = threading.Event()
    failed = threading.Event()

    total_size = 0
    for rsync_file in rsync_files_list.decode("utf-8").strip().split("\n"):
        mode, size, mdate, mtime, filename = re.split(r"\s+", rsync_file, 4)
        if not mode.startswith("d"):
            size = int(size.replace(",", ""))
            download_queue.put(S3Upload(filesize=size, filename=filename))
            total_size += size
    logger.info(
        f"Total component size is {total_size} B - {total_size / (1024*1024*1024):.3f} GB - {download_queue.qsize()} files to sync."
    )

    create_bucket_if_not_exists(s3_config)

    payload = {
        "bucket": s3_config["bucket"],
        "dci_files_list": {"directories": [], "files": []},
        "http": urllib3.PoolManager(maxsize=20),
        "compose_url": component.url,
        "prefix": str(component.id),
    }

    download_threads = [
        threading.Thread(
            target=_get_http_file_worker,
            args=[
                s3_config,
                payload,
                download_queue,
                upload_queue,
                download_done,
                failed,
            ],
            name=f"downloader-{n}",
            daemon=False,
        )
        for n in range(2)
    ]
    upload_threads = [
        threading.Thread(
            target=_push_to_s3_worker,
            args=[
                s3_config,
                payload,
                upload_queue,
                upload_done,
                failed,
            ],
            name=f"uploader-{n}",
            daemon=False,
        )
        for n in range(2)
    ]
    for t in download_threads + upload_threads:
        t.start()

    download_queue.join()
    download_done.set()
    upload_queue.join()
    upload_done.set()

    map(lambda t: t.join(), download_threads + upload_threads)

    if failed.is_set():
        raise Exception("Failed to mirror some files")

    dci_files_list = payload["dci_files_list"]
    create_s3_client(s3_config).put_object(
        Bucket=payload["bucket"],
        Key=os.path.join(payload["prefix"], "dci_files_list.json"),
        Body=json.dumps(dci_files_list),
        ChecksumAlgorithm="sha256",
    )
    logger.info(f"Sync to S3 done - {component.display_name} - {component.id}")
    return dci_files_list


def get_s3_config():
    return {
        "bucket": settings.COMPONENTS_S3_BUCKET,
        "aws_access_key_id": settings.AWS_ACCESS_KEY_ID,
        "aws_secret_access_key": settings.AWS_SECRET_ACCESS_KEY,
        "aws_region": settings.AWS_REGION,
        "aws_endpoint_url": settings.S3_ENDPOINT_URL,
        "aws_signature_version": settings.AWS_SIGNATURE_VERSION,
    }


def sync_compose(component):
    s3_config = get_s3_config()
    dci_files_list = sync_compose_on_s3(s3_config, component)
    component.update_metadata(dci_files_list)


def add_images_list(component, images_list_content):
    s3_config = get_s3_config()
    create_s3_client(s3_config).put_object(
        Bucket=s3_config["bucket"],
        Key=os.path.join(str(component.id), "images_list.yaml"),
        Body=images_list_content,
        ChecksumAlgorithm="sha256",
    )


def list_components_ids():
    s3_config = get_s3_config()
    paginator = create_s3_client(s3_config).get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=s3_config["bucket"], Delimiter="/")
    components_ids = []
    for page in pages:
        for prefix in page["CommonPrefixes"]:
            components_ids.append(prefix["Prefix"].rstrip("/"))
    return components_ids


def remove_from_s3(component):
    s3_config = get_s3_config()
    s3_client = create_s3_client(s3_config)
    bucket = s3_config["bucket"]
    paginator = s3_client.get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=bucket, Prefix=f"{component['id']}/")
    to_delete = {"Objects": []}
    for page in pages:
        for obj in page["Contents"]:
            key = obj["Key"]
            to_delete["Objects"].append({"Key": key})
            logger.debug(f"Deleting file {key}")
            if len(to_delete["Objects"]) >= 1000:
                s3_client.delete_objects(Bucket=bucket, Delete=to_delete)
                to_delete = {"Objects": []}
    if len(to_delete["Objects"]):
        s3_client.delete_objects(Bucket=bucket, Delete=to_delete)
    logger.debug(f"Component {component['id']} removed from s3.")
