# Copyright 2020-2022 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

import logging
import requests
import subprocess
import json

from dcifeeder import settings
from dcifeeder import notify
from dcifeeder.topic import BaseTopic
from dcifeeder.component import BaseComponent
from dcifeeder.exceptions import FailureException
from dcifeeder.feeder import get_or_create_topic_from_name
from dcifeeder.utils.sort import sort_by_oldest_version_first

logger = logging.getLogger(__name__)


def get_builds(release_controler_stream):
    url = (
        "https://amd64.ocp.releases.ci.openshift.org/api/v1/releasestream/%s/tags"
        % release_controler_stream
    )
    logger.debug("Getting builds data from %s" % url)
    try:
        resp = requests.get(url)
        resp.raise_for_status()
        builds = resp.json()["tags"]
    except Exception as e:
        logger.exception(e)
        return []
    if builds:
        return builds
    else:
        return []


def filter_accepted_builds(builds):
    return [b for b in builds if b["phase"].lower() == "accepted"]


def filter_version_builds(builds, ocp_version):
    return [b for b in builds if b["name"].startswith(ocp_version)]


def fix_url(download_url, name, tag):
    if tag == "build:nightly":
        return download_url
    ocp_dir = "ocp"
    if tag == "build:dev":
        ocp_dir = "ocp-dev-preview"
    return "https://mirror.openshift.com/pub/openshift-v4/clients/%s/%s" % (
        ocp_dir,
        name,
    )


def tag_to_sha(image_url, digest):
    if "@" not in image_url and ":" in image_url:
        (base_url, tag) = image_url.split(":")
        image_url = "%s@%s" % (base_url, digest)
    return image_url


def get_release_json_txt(pull_spec):
    logger.info("Getting metadata for %s" % (pull_spec))
    r = subprocess.run(
        [
            "%s" % (settings.OCP_OC_CMD),
            "adm",
            "release",
            "info",
            "--registry-config=%s" % (settings.OCP_PULLSECRET_FILE),
            "--output=json",
            "%s" % (pull_spec),
        ],
        universal_newlines=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if r.returncode != 0:
        raise FailureException(r.stderr)
    logger.info("... success")
    return r.stdout


def get_release_txt(url):
    logger.info(f"Getting release.txt file for {url}")
    r = requests.get(f"{url}/release.txt")
    r.raise_for_status()
    return r.text


def get_openshift_metadata(url, release_json, tags):
    json_data = json.loads(release_json)
    metadata = {}
    digest = json_data["digest"]
    pull_spec = tag_to_sha(json_data["image"], digest)
    metadata["pull_url"] = pull_spec
    version = json_data["metadata"]["version"]
    metadata["version"] = version
    metadata["digest"] = [digest]
    metadata["uid"] = digest
    metadata["created"] = json_data["config"]["created"]
    metadata["display_name"] = "OpenShift %s" % (
        version.split("-")[0] if "-" in version else version
    )
    created = datetime.strptime(metadata["created"], "%Y-%m-%dT%H:%M:%SZ")
    metadata["url"] = url
    # ga and candidate are coming from the same source
    # difference is on version containing rc or fc
    if (
        "build:ga" in tags
        and "version" in metadata
        and ("rc" in metadata["version"] or "fc" in metadata["version"])
    ):
        tags[tags.index("build:ga")] = "build:candidate"
    if "build:candidate" in tags or "build:dev" in tags:
        if "-" in metadata["version"]:
            vers = metadata["version"].split("-")
            metadata["display_name"] = "%s %s" % (
                metadata["display_name"],
                vers[1],
            )
        else:
            metadata["display_name"] = "%s RC" % (metadata["display_name"])
    elif "build:nightly" in tags:
        metadata["display_name"] = "%s nightly %s" % (
            # Remove .0 at the end of the version
            (metadata["display_name"][:-2])
            if metadata["display_name"].endswith(".0")
            else metadata["display_name"],
            # There can be multiple versions a day
            created.strftime("%Y-%m-%d %H:%M"),
        )
    if "build:candidate" in tags:
        # Candidate components may have multiple "releases" with the same name but
        # different "created" date.
        # That's why the "created" date is appended to the component version to
        # differentiate them.
        metadata["version"] += "-%s" % created.strftime("%Y-%m-%d-%H%M%S")
    metadata["tags"] = tags + metadata["digest"]
    return metadata


class OpenshiftComponent(BaseComponent):
    def __init__(self, metadata):
        self.display_name = metadata["display_name"]
        self.version = metadata["version"]
        self.url = metadata["url"]
        self.type = "ocp"
        self.version = metadata["version"]
        self.released_at = metadata["created"]
        super(OpenshiftComponent, self).__init__(self.url)
        self.tags = metadata["tags"]
        self.uid = metadata["uid"]
        self.data = metadata

    def retrieve_metadata(self, metadata_url):
        pass

    def is_healthy(self):
        try:
            if any(
                tag in self.tags for tag in ["build:dev", "build:candidate", "build:ga"]
            ):
                for line in get_release_txt(self.url).split("\n"):
                    if "Version:" in line:
                        version = line.split("Version:")[-1].strip()
                        break
                return version == self.version
            return True
        except Exception:
            return False

    def json(self):
        component = {
            "uid": self.uid,
            "display_name": self.display_name,
            "url": self.url,
            "type": self.type,
            "version": self.version,
            "released_at": self.released_at,
            "data": self.data,
            "tags": self.tags,
        }

        return component


class Openshift(BaseTopic):
    def __init__(self, ocp_version):
        self.name = "OCP-%s" % ocp_version
        super(Openshift, self).__init__()
        self.components = []
        release_controler_streams = {
            "build:ga": "4-stable",
            "build:dev": "4-dev-preview",
            "build:nightly": "%s.0-0.nightly" % ocp_version,
        }
        max_number_of_builds = 5
        for tag, release_controler_stream in release_controler_streams.items():
            releases = filter_version_builds(
                filter_accepted_builds(get_builds(release_controler_stream)),
                ocp_version,
            )[:max_number_of_builds]
            for release in releases:
                url = fix_url(release["downloadURL"], release["name"], tag)
                pull_spec = release["pullSpec"]
                release_json = get_release_json_txt(pull_spec)

                if not release_json:
                    continue
                metadata = get_openshift_metadata(url, release_json, [tag])
                self.components.append(OpenshiftComponent(metadata=metadata))

    def json(self):
        topic = self.topic
        if self.topic:
            topic["components"] = [c.json() for c in self.components]
        return topic

    def get_or_create_topic(self, name):
        topic, created = get_or_create_topic_from_name(name, "OpenShift")
        if created:
            notify.send_message_on_google_chat(
                "Topic %s has been created on DCI https://www.distributed-ci.io/topics/%s/components"
                % (topic["name"], topic["id"])
            )
        return topic


def get_release_streams():
    r = requests.get(
        "https://amd64.ocp.releases.ci.openshift.org/api/v1/releasestreams/all"
    )
    r.raise_for_status()
    return r.json()


def get_supported_openshift_versions():
    release_streams = get_release_streams()
    topic_versions = []
    for release_stream in release_streams.keys():
        parts = release_stream.split(".")
        if len(parts) >= 2:
            topic_version = ".".join(parts[:2])
            if topic_version not in topic_versions:
                topic_versions.append(topic_version)
    return sort_by_oldest_version_first(topic_versions)
