# Copyright 2018 Yanis Guenane  <yanis@guenane.org>
# Author: Yanis Guenane  <yanis@guenane.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bs4
import logging
import re
import requests

from dcifeeder.feeder import get_componment_by_name_type_and_state, update_component
from dcifeeder.components.openstack import OSPFullComposeWithContainers
from dcifeeder.topic import BaseTopic


logger = logging.getLogger(__name__)

OSP_BASE_URL = "http://download.eng.bos.redhat.com/rcm-guest/puddles/OpenStack/"
OSP_RELEASE_FOLDER = "{version}-RHEL-{rhel_ver}/"
OSP_IMPORTED_PHASE = "passed_phase1"

OSP_REPO_URL = (
    OSP_BASE_URL
    + OSP_RELEASE_FOLDER
    + OSP_IMPORTED_PHASE
    + "/RH{rhel_ver}-RHOS-{version}.repo"
)
OSP_COMPOSE_URL = OSP_BASE_URL + OSP_RELEASE_FOLDER + OSP_IMPORTED_PHASE + "/COMPOSE_ID"


class OSP(BaseTopic):
    def component_name_for_tag_url(self, url):
        return requests.get(url + "COMPOSE_ID").text

    def scrape_tags(self, topic_url):
        soup = bs4.BeautifulSoup(requests.get(topic_url).content, "html.parser")
        tags = soup.find_all(
            "a",
            href=re.compile(
                r"^(ga|rc|beta(-\d+(\.\d+)?)?|latest_cdn|passed_phase\d+|z\d+)/$"
            ),
        )
        return {
            tag["href"].rstrip("/"): self.component_name_for_tag_url(
                topic_url + tag["href"]
            )
            for tag in tags
        }

    def update_tags(self):
        topic_url = (OSP_BASE_URL + OSP_RELEASE_FOLDER).format(
            version=self.version, rhel_ver=self.rhel_ver
        )
        logger.info("Updating tags for topic %s from URL %s" % (self.name, topic_url))
        tags_mapping = self.scrape_tags(topic_url)
        logger.debug("Scraped the following tags: %s" % tags_mapping)

        for tag, component_name in tags_mapping.items():
            component = get_componment_by_name_type_and_state(
                topic=self,
                component={
                    "display_name": component_name,
                    "type": self.component_types[0],
                },
                state="active",
            )
            if component and (
                component["tags"] is None or tag not in component["tags"]
            ):
                if component["tags"] is None:
                    component["tags"] = [
                        tag,
                    ]
                else:
                    component["tags"].append(tag)
                logger.info(
                    "Adding tag %s to component %s/%s (%s)"
                    % (tag, self.name, component["display_name"], component["id"])
                )
                component = update_component(component, {"tags": component["tags"]})
            else:
                if not component:
                    logger.debug("No active component for: %s" % component_name)
                elif tag in component["tags"]:
                    logger.debug(
                        "Tag %s already applied to component %s"
                        % (tag, component["name"])
                    )


class OSP16_1(OSP):
    def __init__(self):
        self.name = "OSP16.1"
        super(OSP16_1, self).__init__()
        self.version = "16.1"
        self.rhel_ver = "8"
        self.url = OSP_COMPOSE_URL.format(version=self.version, rhel_ver=self.rhel_ver)
        self.components = [OSPFullComposeWithContainers(metadata_url=self.url)]
        self.component_types = ["compose"]


class OSP16_2(OSP):
    def __init__(self):
        self.name = "OSP16.2"
        super(OSP16_2, self).__init__()
        self.version = "16.2"
        self.rhel_ver = "8"
        self.url = OSP_COMPOSE_URL.format(version=self.version, rhel_ver=self.rhel_ver)
        self.components = [OSPFullComposeWithContainers(metadata_url=self.url)]
        self.component_types = ["compose"]


class OSP17_0(OSP):
    def __init__(self):
        self.name = "OSP17.0"
        super(OSP17_0, self).__init__()
        self.version = "17.0"
        self.rhel_ver = "9"
        self.url = OSP_COMPOSE_URL.format(version=self.version, rhel_ver=self.rhel_ver)
        self.components = [OSPFullComposeWithContainers(metadata_url=self.url)]
        self.component_types = ["compose"]


class OSP17_1(OSP):
    def __init__(self):
        self.name = "OSP17.1"
        super(OSP17_1, self).__init__()
        self.version = "17.1"
        self.rhel_ver = "9"
        self.url = OSP_COMPOSE_URL.format(version=self.version, rhel_ver=self.rhel_ver)
        self.components = [OSPFullComposeWithContainers(metadata_url=self.url)]
        self.component_types = ["compose"]


OPENSTACK_TOPICS_LIST = {
    "osp16.1": OSP16_1,
    "osp16.2": OSP16_2,
    "osp17.0": OSP17_0,
    "osp17.1": OSP17_1,
}
